/****************************************************************************
   File: EgBase.h
   Created: 8/7/2001
   
   Author: Tim Halverson, Erik Brown
   Principle Investigator: Anthony Hornof
   Cognitive Modeling and Eye Tracking Lab
   CIS Department
   University of Oregon
   
   This is a modification of:
      EgWin.h       Windows interface functions to the Eyegaze System
      
      Created 11/11/1998, LC Technologies, Inc.

		Description: EgBase and its derivatives, EgMac and EgWin,
		provide a connection to an LC Technologies eye tracking
		system. An object created can sent four messages to the eye
		tracker: connect, start data, stop data, and disconnect. The
		object will also collect eye data that the server sends, and
		provides a public method that the owner of the Eg object uses
		to get this data.

		Modifications:

		2007.8.1 - eb
		The original code provided much more functionality than it does now.
		No calibration code is in place any longer. Only one message is processed
		from the eye tracker -- a gaze info message. UNIX sockets and POSIX threads
		are used in place of Mac dependent APIs. Logging functions have been left
		in place, but I don't know whether they still work.

		2007.10.3 - eb
		Logging functions have been removed, but I haven't even begun to start!
		The code has been drastically modified so that it will also build for the
		Windows platform. You are now looking at an abstract base class that provides
		most of the functionality, everything that is platform independent. The user
		will actually instantiate an object derived from this, either an EgMac or EgWin
		object. See docs for more information.

   Copyright (c) 2007 by the University of Oregon.
   ALL RIGHTS RESERVED.

   Permission to use, copy, and distribute this software in
   its entirety for non-commercial purposes and without fee,
   is hereby granted, provided that the above copyright notice
   and this permission notice appear in all copies and their
   documentation.

   Software developers, consultants, or anyone else who wishes
   to use all or part of the software or its documentation for
   commercial purposes should contact the Technology Transfer
   Office at the University of Oregon to arrange a commercial
   license agreement.

   This software is provided "as is" without expressed or
   implied warranty of any kind.
*****************************************************************************/

#ifndef EGBASE_H
#define EGBASE_H

#include <deque>

#ifdef WIN_VERSION
	#include <windows.h>
	typedef unsigned long UInt32;
#elif MAC_VERSION
	#include <sys/types.h>
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <netdb.h>
#endif

#include "EGStructs.h"
#include "EgConstants.h"

typedef enum 
{
	EG_NOERROR, 
	EG_ERROR, 
	EG_CONNECTIONFAILED, 
	EG_RECEIVEFAILED
} EgStatus;

typedef enum 
{
	STATE_START, 
	STATE_CONNECTED, 
	STATE_RECEIVING
} EgState;

//////////////////////////////////////////////////////////////////////////
//
// Routine to read from open socket connected to eye tracker. Different 
// platforms require a different function signature, but the code inside
// is platform independent.
//
#ifdef WIN_VERSION
	#include <windows.h>
	DWORD WINAPI EyeTrackerReceive(void* td);
#elif MAC_VERSION
	void* EyeTrackerReceive(void* td);
#endif


class EgBase
{
public:
	static const int numDataPoints;  // how many gaze points to keep in the queue
                                    // refer to the documentation for info
   
public:
	EgBase();
	virtual ~EgBase();

	// Need to be interface methods. The derived classes use similar, but
	// platform specific, ways of reading and writing sockets.
	virtual void   writeSocket(const char *buffer, int len) = 0;
	virtual int    readSocket(char *buffer, int bytesToRead) = 0;

	
	// Called from the application, the owner of the Eg object.
	EgStatus	   connect(const char *commName);
	EgStatus		disconnect();
	EgStatus		startData();
	EgStatus		stopData();

	bool	isConnected();

	bool GetData(stEgData &data);
	void SetData(stEgData &data);

protected:
	int theSocket;
	
private:    
	void sendMessage	(int iMessage, unsigned char *pszBuffer, unsigned int uiBufferLen);
	
	// Need to be interface methods. The derived classes ensure exclusive
	// access, create threads, etc., in platform specific ways.
	virtual void		startThread() = 0;
	virtual void      closeSocket() = 0;
	virtual void      Sleep(unsigned long milliseconds) = 0;
	
	virtual void		WaitMutex() = 0;
	virtual void		SignalMutex() = 0;

	std::deque<stEgData>	egDataDeque;

	EgState				currentState;
};

#endif // EGBASE_H
