 ///////////////////////////////////////////////////////////////////
 //
 // File: main.cpp
 // 
 // Project: fixation Max external.
 // 
 // Authors: Erik Brown, Linda Sato.
 // Principle Investigator: Anthony Hornof
 // Cognitive Modeling and Eye Tracking Lab
 // CIS Department, University of Oregon
 // 
 // Description: Provides a Max external that accepts eye data input
 //					and outputs fixations in real time.
 //
 //					Input consists of a continuous stream of data from
 //					the 'eyedata' object: the x and y positions and the
 //					eyefound flag. Additional inputs specify the sample
 //					rate from the eyetracker, the fixation algorithm, and
 //					the parameters (threshold, min time) for the algorithms.
 //
 //					Outputs x and y positions when a new fixation occurs, also
 //					when a fixation begins and ends.
 //
 // 
 // Copyright (c) 2007 by the University of Oregon.
 // ALL RIGHTS RESERVED.
 //
 // 
 // Permission to use, copy, and distribute this software in
 // its entirety for non-commercial purposes and without fee,
 // is hereby granted, provided that the above copyright notice
 // and this permission notice appear in all copies and their
 // documentation.
 // 
 // Software developers, consultants, or anyone else who wishes
 // to use all or part of the software or its documentation for
 // commercial purposes should contact the Technology Transfer
 // Office at the University of Oregon to arrange a commercial
 // license agreement.
 // 
 // This software is provided "as is" without expressed or
 // implied warranty of any kind.
 // 
 //
 ///////////////////////////////////////////////////////////////////
 
 
#include <math.h>

#include "ext.h"
#include "ext_obex.h"
#include "velocity.h"

short MOVING = 0;
short FIXATING = 1;
short FIXATION_COMPLETED = 2;

typedef enum {DISPERSION_ALG, VELOCITY_ALG} FixationAlgorithm;
                                       
/* structure definition of your object */
typedef struct fixation
{
	t_object f_ob;
	void *obex;
	void *f4_out;				// end fixation outlet
	void *f3_out;				// begin fixation outlet
	void *f2_out;				// y position of current fixation beginning
	void *f1_out;				// x position of current fixation beginning
	long f_xvalue;				// incoming x coord
	long f_yvalue;				// incoming y coord

	FixationAlgorithm	algorithm;			// which fixation algorithm to use
	int	sample_rate;
	int	min_fix_time;
	float	dis_threshold;
	float vel_threshold;
	
	VF::Velocity *velocityFixation;
} t_fixation;

#define RING_SIZE              31      /* length of the delay line in       */
                                       /*   DetectFixation() --             */
                                       /*   should be greater than          */
                                       /*   minimum_fix_samples             */

/* globalthat holds the class definition */
t_class *fixation_class;

// Prototypes. Declared as extern "C" to prevent name mangling.
extern "C"{
	int	main				(void);
	void* fixation_new	(long n);
	void fixation_delete	(t_fixation *x);

	void inX				(t_fixation *x, long n);	//takes y coord
	void inY				(t_fixation *x, long n);	//takes x coord
	void inEyeFound	(t_fixation *x, long n);	//takes eye found flag
	void inSampleRate	(t_fixation *x, long n);	//takes a sample rate
	void inAlgorithm	(t_fixation *x, t_symbol *alg);	//takes an algorithm string
	void inThreshold	(t_fixation *x, double n);	//takes a fixation threshold
	void inMinTime		(t_fixation *x, long n);	//takes a minimum time for a fixation
		
		//void fixation_bang(t_fixation *x);
	//  void fixation_set(t_fixation *x, long n);
	void reset_mess(t_fixation *x);//to call the initializer
	void fixation_assist(t_fixation *x,  void *b, long m, long a, char *s);
	void scopy(char *s1, char *s2);
	void fixation_anything(t_fixation *x, t_symbol *message, short argc, t_atom *argv);
	/* FUNCTION PROTOTYPES from fixfunc:                                                     */
}

void ResetPresFixation(void);
void ResetNewFixation(void);
void StartPresFixAtGazepoint(float x_gaze, float y_gaze);
void StartNewFixAtGazepoint(float x_gaze, float y_gaze);
void UpdatePresFixation(float x_gaze, float y_gaze, short minimum_fix_samples);
void UpdateNewFixation(float x_gaze, float y_gaze);
void CalcGazeDeviationFromPresFix(float x_gaze, float y_gaze);
void CalcGazeDeviationFromNewFix(float x_gaze, float y_gaze);
void CheckIfFixating(short minimum_fix_samples);
void MoveNewFixToPresFix(short minimum_fix_samples);
void DeclareCompletedFixation(short minimum_fix_samples);
void RestoreOutPoints(void);
void InitFixation(short minimum_fix_samples);
short DetectFixation(Boolean  gazepoint_found, float x_gaze, float y_gaze, 
	float gaze_deviation_threshold, short minimum_fix_samples, Boolean  *ptr_gazepoint_found_delayed, 
	float *ptr_x_gaze_delayed, float *ptr_y_gaze_delayed, float *ptr_gaze_deviation_delayed,  
	float *ptr_x_fix_delayed, float *ptr_y_fix_delayed, short *ptr_saccade_duration_delayed, 
	short *ptr_fix_duration_delayed);//wow, that's a lot of args
void CMET_Debug(char *message);
	
//Functions added to FixFunc
void resetState(void);

/* GLOBAL FIXFUNC VARIABLES:                                                */

long  lCallCount;              /* number of times this function has been   */
                                /*   called since it was initialized        */
                                /*   (30ths or 60ths of a second,           */
                                /*   depending on eyetracking sample rate)  */
short   iNNoEyeFound;            /* number of successive samples with no     */
                                /*   eye found                              */


                                /* DATA ON PREVIOUS FIXATION                */
long  lPrevFixEndCount;        /* count that the previous fixation ended   */

                                /* DATA ON PRESENT FIXATION                 */
long  lPresFixStartCount;      /* call count that the fixation starts      */
long  lPresFixEndCount;        /* call count that the fixation ends        */
short   iNPresFixSamples;        /* number of samples in the present fix     */
float fXPresFixSum;            /* summations for calculation of average    */
float fYPresFixSum;            /*   fixation position                      */
float fXPresFix;               /* average coordinate of the eye fixation   */
float fYPresFix;               /*   point (user selected units)            */
short   iNPresOut;               /* number of samples outside the fixation   */
float fPresDr;                 /* difference between gazepoint and         */
                                /*   fixation (x, y, and radius)            */

                                /* DATA ON NEW FIXATION                     */
long  new_fix_start_count;     /* call count that the new fixation starts  */
long  new_fix_end_count;       /* call count that the new fixation ends    */
short   n_new_fix_samples;       /* number of samples in the fixation        */
float x_new_fix_sum;           /* summations for the FIR filter            */
float y_new_fix_sum;           /*   calculations of the eye motion         */
float x_new_fix;               /* average coordinate of the eye fixation   */
float y_new_fix;               /*   point (user selected units)            */
float new_dr;                  /* difference between gazepoint and         */
                                /*   fixation (x, y, and radius)            */

                                /* RING BUFFERS STORING PAST VALUES:        */
float x_gaze_ring[RING_SIZE];
float y_gaze_ring[RING_SIZE];
Boolean gaze_found_ring[RING_SIZE];
short   eye_motion_state[RING_SIZE];
                                /* state of the eye motion:                 */
                                /*   MOVING                                 */
                                /*   FIXATING                               */
                                /*   FIXATION_COMPLETED                     */
float x_fix_ring[RING_SIZE];
float y_fix_ring[RING_SIZE];
float gaze_deviation_ring[RING_SIZE];
short   sac_duration_ring[RING_SIZE];
short   fix_duration_ring[RING_SIZE];
short   iRingIndex;              /* ring index of the present gaze sample    */
short   iRingIndexDelay;        /* ring index of the gaze sample taken      */
                                /*   minimum_fix_samples ago                */


/*****global variables added by Linda**********/
Boolean beginning_of_fix;//initialize to TRUE in initializer
float mygaze_deviation_threshold;// initialize to 1 in main, then set by user
short myminimum_fix_samples;//initialize to 6 in main, then set by user
long xcoord;  
short eyefoundflag;

//the following are all args in the call to detectfix, 
//updated after call to detectfix
Boolean *myptr_gazepoint_found_delayed;
float *myptr_x_gaze_delayed;
float *myptr_y_gaze_delayed;
float *myptr_gaze_deviation_delayed;
float *myptr_x_fix_delayed;
float *myptr_y_fix_delayed;
short *myptr_saccade_duration_delayed;
short *myptr_fix_duration_delayed;
//the above pointers are pointed at this stuff in main
Boolean init_gazepoint_found_delayed;
float init_x_gaze_delayed;
float init_y_gaze_delayed;
float init_gaze_deviation_delayed;
float init_x_fix_delayed;
float init_y_fix_delayed;
short init_saccade_duration_delayed;
short init_fix_duration_delayed;

/* initialization routine for Max*/

int main(void)
{
	/* tell Max about your class. */
   t_class *c;
	
   c = class_new("fixation", (method)fixation_new, (method)fixation_delete,
					  (short)sizeof(t_fixation), (method)0L, 0);
	
	class_obexoffset_set(c, calcoffset(t_fixation, obex));
	
   common_symbols_init();
	
   //class_addmethod(c, (method)reset_mess, "reset", A_NOTHING);
   class_addmethod(c, (method)fixation_assist, "assist", A_CANT, A_NOTHING);
   class_addmethod(c, (method)fixation_anything, "anything", A_GIMME, A_NOTHING);
	
	/////////////////////////////////////////////////////////////////////
	// Our inlets.
	//
	// x pos
	class_addmethod(c, (method)inX, "int", A_LONG, A_NOTHING);
	
	// y pos
	class_addmethod(c, (method)inY, "in1", A_LONG, A_NOTHING);
	
	// eye found flag
	class_addmethod(c, (method)inEyeFound, "in2", A_LONG, A_NOTHING);
	
	// sample rate
	class_addmethod(c, (method)inSampleRate, "in3", A_LONG, A_NOTHING);
	
	// fixation algorithm
	class_addmethod(c, (method)inAlgorithm, "algorithm", A_DEFSYM, A_NOTHING);
	
	// fixation threshold
	class_addmethod(c, (method)inThreshold, "ft5", A_FLOAT, A_NOTHING);
	
	// fixation minimum time
	class_addmethod(c, (method)inMinTime, "in6", A_LONG, A_NOTHING);
	//
	/////////////////////////////////////////////////////////////////////
		
	class_register(CLASS_BOX, c);
	fixation_class = c;
	
	//attach all the pointers to stuff
	myptr_gazepoint_found_delayed = &init_gazepoint_found_delayed;
 	myptr_x_gaze_delayed = &init_x_gaze_delayed;
	myptr_y_gaze_delayed = &init_y_gaze_delayed;
 	myptr_gaze_deviation_delayed = &init_gaze_deviation_delayed;
 	myptr_x_fix_delayed = &init_x_fix_delayed;
 	myptr_y_fix_delayed = &init_y_fix_delayed;
 	myptr_saccade_duration_delayed = &init_saccade_duration_delayed;
 	myptr_fix_duration_delayed = &init_fix_duration_delayed;
	mygaze_deviation_threshold = 20.0;//default value, can be changed by user
	myminimum_fix_samples = 6;//default value, can be changed by user
		
	return 0;
}

/* instance creation routine */

void *fixation_new(long n)
{
	t_fixation *x;
	
	x = (t_fixation*)object_alloc(fixation_class);
	
	x->f_xvalue = n;						// store value (default is 0)
	x->f1_out = intout(x);				// create an int outlet
	x->f2_out = floatout(x);				// create an float outlet
	x->f3_out = floatout(x);				// create an float outlet
	
	// Make our inlets. Since the x pos is the first inlet, and automatically
	// available to, we don't need to create it.
	intin(x, 6);
	floatin(x, 5);
	inlet_new(x, 0);
	intin(x, 3);
	intin(x, 2);
	intin(x, 1);
	
	x->algorithm = DISPERSION_ALG;
	x->sample_rate = 60; // default, can be changed by user
	x->min_fix_time = 200;
	x->dis_threshold = 20.0;
	x->vel_threshold = 20.0;
	
	x->velocityFixation = new VF::Velocity;
	
	return (x);							// return newly created object to caller
}

void fixation_delete(t_fixation *x)
{
	if (NULL != x->velocityFixation)
	{
		delete x->velocityFixation;
		x->velocityFixation = NULL;
	}
}

void fixation_assist(t_fixation *x, void *b, long m, long a, char *s)
{
	if (m == 1) //i.e. if message passed is inlet
	{
		switch (a)
		{
			case 0: 
				scopy (s, "int, x coord.");
				break;
				
			case 1: 
				scopy (s, "int, y coord.");
				break;
				
			case 2: 
				scopy (s, "int, eyefound flag");
				break;
				
			case 3: 
				scopy(s, "int, sample rate from eye tracker");
				break;
				
			case 4:
				scopy(s, "message, fixation algorithm");
				break;
				
			case 5:
				scopy(s, "float, fixation threshold");
				break;
				
			case 6:
				scopy(s, "int, min time(ms) for fixation");
				break;
		}
	}
	else if (m == 2) //i.e. if message passed is outlet
	{
		switch (a)
		{
			case 0:	
				scopy (s, "float, fix x coord");
				break;

			case 1: 
				scopy (s, "float, fix y coord");
				break;
				
			case 2: 
				scopy (s, "int, 0 when fix has ended");
				break;
		}
	}
}
void scopy(char *s1, char *s2)
{
	short i;
	i = 0;
	while ((s1[i] = s2[i]) != '\0')
		i++;

} 

void inX(t_fixation *x, long n)//store x
{

	x->f_xvalue = n;
	xcoord = n;
	//post("should be x: %ld", (long) n);
							// store a value inside the object
	//outlet_int(x->f_out,x->f_value);	// and send it out
}
void inY(t_fixation *x, long n)//store y
{
	
	//the y coord is the last necessary piece of info for calling detectFix, so the call happens
	//here.
	Boolean mygazepoint_found;
	short result;
	long ix, iy;
	float fx, fy;
	x->f_yvalue = n;
	//ix = x->f_xvalue;
	ix = xcoord;
	iy = n;
	fx = (float)ix;
	fy = (float)iy;
	
							// store a value inside the object
	//if (x->f_eyefoundvalue == 1)
	if(eyefoundflag == 1)
		mygazepoint_found = TRUE;
	else 
		mygazepoint_found = FALSE;
		//post("eyefound: %ld", eyefoundflag);
	//post("should be y: %ld", (long) n);
	//post("here is x: %ld", (long) ix);
	//post("here is y: %ld", (long) iy);
	
	//post("here is x: %lf", (float) fx);
	//post("here is y: %lf", (float) fy);
	//note to self: I tried to pass the pointers with their *'s but that don't work!!
	
	
	
	switch (x->algorithm)
	{
		case DISPERSION_ALG:
			{
				result = DetectFixation(mygazepoint_found, fx, fy, mygaze_deviation_threshold, myminimum_fix_samples, 
						myptr_gazepoint_found_delayed, myptr_x_gaze_delayed, myptr_y_gaze_delayed, myptr_gaze_deviation_delayed,
						myptr_x_fix_delayed, myptr_y_fix_delayed, myptr_saccade_duration_delayed, myptr_fix_duration_delayed);

				if (result == 1){
					if (beginning_of_fix == TRUE){
						beginning_of_fix = FALSE;
						fx = *myptr_x_fix_delayed;
						fy = *myptr_y_fix_delayed;
						outlet_float(x->f3_out, fx);//output x and y
						outlet_float(x->f2_out, fy);
					}
				}
				else if (result == 2){
					outlet_int(x->f1_out, 0);//output the stop message
					beginning_of_fix = TRUE;
					resetState();//this goes back and changes the state of the last minfixsamples to MOVING
				}
			}

			break;
			
		case VELOCITY_ALG:
			{
				// We are assuming here that the eyefound flag, retrieved in a separate
				// function, goes along with this (x and) y data, and is not the
				// eye found flag of the previous sample. In other words, data comes to
				// us in this order: eyefound flag before y position.

				VF::Point pt = { ix, iy };
				VF::FixationState state = x->velocityFixation->Add(pt, eyefoundflag);

				switch(state)
				{
					case VF::FIXBEGIN:
						outlet_float(x->f3_out, fx);
						outlet_float(x->f2_out, fy);
						break;

					case VF::FIXEND:
						outlet_int(x->f1_out, 0);
						break;

					default:
						break;	// VF::MOVING, VF::FIX
				}
			}
			break;
			
		default:
			break;
	}
}

void inEyeFound(t_fixation *x, long n)//store eye found flag and call detectfix
{

	eyefoundflag = n;
	//x->f_eyefoundvalue = n;
	//post("should be eyefound: %ld", (long) n);
		
			 
	//outlet_int(x->f_out,x->f_value);	// and send it out
	
}

void inAlgorithm	(t_fixation *x, t_symbol *s)
{
	// Only switch algorithms if we are given a different algorithm
	// than the current.
	if (strncmp("velocity", s->s_name, strlen("velocity")) == 0)
	{
		if (x->algorithm != VELOCITY_ALG)
		{
			x->algorithm = VELOCITY_ALG;
			
			// Make sure to give the velocity object a "good" starting point.
			VF::Point xy = {x->f_xvalue,x->f_yvalue};
			x->velocityFixation->SetCurrentPoint(xy);
			
			// Run through these to pick up currently set values in the patch.
			x->velocityFixation->SetSampleRate(x->sample_rate);
			x->velocityFixation->SetThreshold(x->vel_threshold);
			x->velocityFixation->SetMinFixTime(x->min_fix_time);
		}
	}
	else if (strncmp("dispersion", s->s_name, strlen("dispersion")) == 0)
	{
		if (x->algorithm != DISPERSION_ALG)
		{
			x->algorithm = DISPERSION_ALG;

			myminimum_fix_samples = ceil(x->min_fix_time / (1000.0 / x->sample_rate));
			mygaze_deviation_threshold = x->dis_threshold;
		}
	}
	
	
	char buf[512];
	sprintf(buf, "Using %s based detection.\n sample rate: %d\n min fix time: %d\n threshold: %f",
			  (x->algorithm == VELOCITY_ALG) ? "velocity" : "dispersion", 
			  x->sample_rate, 
			  x->min_fix_time, 
			  (x->algorithm == VELOCITY_ALG) ? x->vel_threshold : x->dis_threshold);
	
	CMET_Debug(buf);
	
	return;
}

void inSampleRate	(t_fixation *x, long n)
{
	x->sample_rate = n;

	x->velocityFixation->SetSampleRate(n);
	myminimum_fix_samples = ceil(x->min_fix_time / (1000.0 / x->sample_rate));
}

void inThreshold(t_fixation *x, double n)
{
	// We've got a "situation" here. Each algorithm's threshold is measured
	// differently. Pixels vs. degrees/second. This could be trouble eventually!
	
	// What we'll do here is to change the setting only for the current
	// algorithm. Be aware, changing this for the current algorithm, then
	// switching algorithms will result in the newly chosen algorithm using
	// it's own previous value. Therefore, to make this explicit to the
	// Max user, he or she ought to have two objects, one to set the value
	// for dispersion and one to set the value for velocity.

	switch(x->algorithm)
	{
		case VELOCITY_ALG:
			x->vel_threshold = n;
			x->velocityFixation->SetThreshold(n);
			break;
			
		case DISPERSION_ALG:
			x->dis_threshold = n;
			mygaze_deviation_threshold = n;
			break;
			
		default:
			break;
	}
}

void inMinTime (t_fixation *x, long n)
{
	x->min_fix_time = n;

	x->velocityFixation->SetMinFixTime(n);
	myminimum_fix_samples = ceil(x->min_fix_time / (1000.0 / x->sample_rate));
}

void reset_mess(t_fixation *x)
{
	InitFixation(6);//acc to notes, this should be called before calling DetectFixation
					//arg is int minimum_fix_samples; used 6  per Karsh and Breitenbach
}

void CMET_Debug(char *message)
{
#ifdef CMET_DEBUG
	post(message);
#endif // CMET_DEBUG
}

//void myobject_set(MyObject *x, long n)
//{
//	x->m_value = n;		// just store incoming value without outputting it
//}
void fixation_anything (t_fixation *x, t_symbol *message, short argc, t_atom *argv)
{
	switch (argv[0].a_type) 
	{
		case A_LONG:
			myminimum_fix_samples = argv[0].a_w.w_long;
			post ("minfix: %ld", (short) myminimum_fix_samples);
			break;
			
		case A_FLOAT:
			mygaze_deviation_threshold = argv[0].a_w.w_float;
			post ("gazedev: %lf", (float)mygaze_deviation_threshold);
			break;
	}
	//else post("message not understood.");
}


	/****************************************************************************/
void InitFixation(short minimum_fix_samples)
                                       /* minimum number of gaze samples    */
                                       /*   that can be considered a        */
                                       /*   fixation                        */
                                       /*   Note: if the input value is     */
                                       /*   is less than 3, the function    */
                                       /*   sets it to 3                    */

/* This function clears any previous, present and new fixations, and it     */
/* initializes DetectFixation()'s internal ring buffers of prior            */
/* gazepoint data.  InitFixation() should be called prior to a sequence     */
/* of calls to DetectFixation().                                            */

{
/* Initialize the internal ring buffer.                                     */
   for (iRingIndex = 0; iRingIndex < RING_SIZE; iRingIndex++)
   {
      x_gaze_ring[iRingIndex]         =  0.0F;
      y_gaze_ring[iRingIndex]         =  0.0F;
      gaze_found_ring[iRingIndex]     =  FALSE;
      eye_motion_state[iRingIndex]    =  MOVING;
      x_fix_ring[iRingIndex]          =  0.0F;
      y_fix_ring[iRingIndex]          =  0.0F;
      gaze_deviation_ring[iRingIndex] = -0.1F;
      sac_duration_ring[iRingIndex]   =  0;
      fix_duration_ring[iRingIndex]   =  0;
   }
   iRingIndex = 0;
   iRingIndexDelay = RING_SIZE - minimum_fix_samples;

/* Set the call count to zero, and initialize the previous fixation end     */
/* count so the first saccade duration is a legitimate count.               */
   lCallCount = 0;
   lPrevFixEndCount = 0;

/* Reset the present fixation data.                                         */
   ResetPresFixation();

/* Reset the new fixation data.                                             */
   ResetNewFixation();

/* Initialize the number of successive samples with no eye found.           */
   iNNoEyeFound = 0;
   
   beginning_of_fix = TRUE;//Linda added this
   post("init called");
}


/****************************************************************************/
short DetectFixation(
                                       /* INPUT PARAMETERS:                 */
       Boolean gazepoint_found,          /* flag indicating whether or not    */
                                       /*   the image processing algo       */
                                       /*   detected the eye and computed   */
                                       /*   a valid gazepoint (TRUE/FALSE)  */
       float x_gaze,                   /* present gazepoint                 */
       float y_gaze,                   /*   (user specified units)          */
       float gaze_deviation_threshold, /* distance that a gazepoint may     */
                                       /*   vary from the average fixation  */
                                       /*   point and still be considered   */
                                       /*   part of the fixation            */
                                       /*   (user specified units)          */
       short   minimum_fix_samples,      /* minimum number of gaze samples    */
                                       /*   that can be considered a        */
                                       /*   fixation                        */
                                       /*   Note: if the input value is     */
                                       /*   is less than 3, the function    */
                                       /*   sets it to 3                    */

                                       /* OUTPUT PARAMETERS:                */
                                       /* Delayed Gazepoint data with       */
                                       /*   fixation annotations:           */
       Boolean  *ptr_gazepoint_found_delayed,
                                       /* sample gazepoint-found flag,      */
                                       /*   min_fix_samples ago             */
       float *ptr_x_gaze_delayed,      /* sample gazepoint coordinates,     */
       float *ptr_y_gaze_delayed,      /*   min_fix_samples ago             */
       float *ptr_gaze_deviation_delayed,
                                       /* deviation of the gaze from the    */
                                       /*   present fixation,               */
                                       /*   min_fix_samples ago             */

                                       /* t_fixation data - delayed:          */
       float *ptr_x_fix_delayed,       /* fixation point as estimated       */
       float *ptr_y_fix_delayed,       /*   min_fix_samples ago             */
       short   *ptr_saccade_duration_delayed,
                                       /* duration of the saccade           */
                                       /*   preceeding the preset fixation  */
                                       /*   (samples)                       */
       short   *ptr_fix_duration_delayed)/* duration of the present fixation  */
                                       /*   (samples)                       */

/* RETURN VALUES - Eye Motion State:                                        */
/*                                                                          */
/*  MOVING                0   The eye was in motion min_fix_samples ago.    */
/*  FIXATING              1   The eye was fixating min_fix_samples ago.     */
/*  FIXATION_COMPLETED    2   A completed fixation has just been detected;  */
/*                              the fixation ended min_fix_samples ago.     */
/*                                                                          */
/* Include FIXFUNC.H for function prototype and above constant definitions. */
/*                                                                          */
/* SUMMARY                                                                  */
/*                                                                          */
/*    This function converts a series of uniformly-sampled (raw) gaze       */
/* points into a series of variable-duration saccades and fixations.        */
/* t_fixation analysis may be performed in real time or after the fact.  To   */
/* allow eye fixation analysis during real-time eyegaze data collection,    */
/* the function is designed to be called once per sample.  When the eye     */
/* is in motion, ie during saccades, the function returns 0 (MOVING).       */
/* When the eye is still, ie during fixations, the function returns 1       */
/* (FIXATING).  Upon the detected completion of a fixation, the function    */
/* returns 2 (FIXATION_COMPLETED) and produces:                             */
/*   a) the time duration of the saccade between the last and present       */
/*      eye fixation (eyegaze samples)                                      */
/*   b) the time duration of the present, just completed fixation           */
/*      (eyegaze samples)                                                   */
/*   c) the average x and y coordinates of the eye fixation                 */
/*      (in user defined units of x_gaze and y_gaze)                        */
/* Note: Although this function is intended to work in "real time", there   */
/* is a delay of minimum_fix_samples in the filter which detects the        */
/* motion/fixation condition of the eye.                                    */
/*                                                                          */
/* PRINCIPLE OF OPERATION                                                   */
/*                                                                          */
/*    This function detects fixations by looking for sequences of gaze-     */
/* point measurements that remain relatively constant.  If a new gazepoint  */
/* lies within a circular region around the running average of an on-going  */
/* fixation, the fixation is extended to include the new gazepoint.         */
/* (The radius of the acceptance circle is user specified by setting the    */
/* value of the function argument gaze_deviation_threshold.)                */
/*    To accommodate noisy eyegaze measurements, a gazepoint that exceeds   */
/* the deviation threshold is included in an on-going fixation if the       */
/* subsequent gazepoint returns to a position within the threshold.         */
/*    If a gazepoint is not found, during a blink for example, a fixation   */
/* is extended if a) the next legitimate gazepoint measurement falls within */
/* the acceptance circle, and b) there are less than minimum_fix_samples    */
/* of successive missed gazepoints.  Otherwise, the previous fixation       */
/* is considered to end at the last good gazepoint measurement.             */
/*                                                                          */
/* UNITS OF MEASURE                                                         */
/*                                                                          */
/*    The gaze position/direction may be expressed in any units (e.g.       */
/* millimeters, pixels, or radians), but the filter threshold must be       */
/* expressed in the same units.                                             */
/*                                                                          */
/* INITIALIZING THE FUNCTION                                                */
/*                                                                          */
/*    Prior to analyzing a sequence of gazepoint data, the InitFixation     */
/* function should be called to clear any previous, present and new         */
/* fixations and to initialize the ring buffers of prior gazepoint data.    */
/*                                                                          */
/* PROGRAM NOTES                                                            */
/*                                                                          */
/* For purposes of describing an ongoing sequence of fixations, fixations   */
/* in this program are referred to as "previous", "present", and "new".     */
/* The present fixation is the one that is going on right now, or, if a     */
/* new fixation has just started, the present fixation is the one that      */
/* just finished.  The previous fixation is the one immediatly preceeding   */
/* the present one, and a new fixation is the one immediately following     */
/* the present one.  Once the present fixation is declared to be completed, */
/* the present fixation becomes the previous one, the new fixation becomes  */
/* the present one, and there is not yet a new fixation.                    */

/*--------------------------------------------------------------------------*/
{
/* Make sure the minimum fix time is at least 3 samples.                    */
   if (minimum_fix_samples < 3) minimum_fix_samples = 3;

/* Make sure the ring size is large enough to handle the delay.             */
   if (minimum_fix_samples >= RING_SIZE)
   {
     // lct_settextmode();
     post("minimum_fix_samples %i >= RING_SIZE %i\n",
              minimum_fix_samples,      RING_SIZE);//post is to Max mess. window
     // printf("Press any key to terminate...");
     // getch();
      //exit(99);
      return -1;//try this instead of the exit(99)
   }

/* Increment the call count, the ring index, and the delayed ring index.    */
   lCallCount++;
   iRingIndex++;
   if (iRingIndex >= RING_SIZE) iRingIndex = 0;
   iRingIndexDelay = iRingIndex - minimum_fix_samples;
   if (iRingIndexDelay < 0) iRingIndexDelay += RING_SIZE;

/* Update the storage rings.                                                */
   x_gaze_ring[iRingIndex]     = x_gaze;
   y_gaze_ring[iRingIndex]     = y_gaze;
   gaze_found_ring[iRingIndex] = gazepoint_found;

/* Initially assume the eye is moving.                                      */
/* Note: These values are updated during the processing of this and         */
/* subsequent gazepoints.                                                   */
   eye_motion_state[iRingIndex]    = MOVING;
   x_fix_ring[iRingIndex]          = -0.0F;
   y_fix_ring[iRingIndex]          = -0.0F;
   gaze_deviation_ring[iRingIndex] = -0.1F;
   sac_duration_ring[iRingIndex]   =  0;
   fix_duration_ring[iRingIndex]   =  0;

/*- - - - - - - - - - - - - Process Tracked Eye  - - - - - - - - - - - - - -*/

/* A1 If the eye's gazepoint was successfully measured this sample,         */
   if (gazepoint_found == TRUE)
   {
/*    The number of successive no-tracks is zero.                           */
      iNNoEyeFound = 0;

/* B1 If there is a present fixation,                                       */
      if (iNPresFixSamples > 0)
      {
/*       Compute the deviation of the gazepoint from the present fixation.  */
         CalcGazeDeviationFromPresFix(x_gaze, y_gaze);

/*   C1  If the gazepoint is within the present fixation region,            */
         if (fPresDr <= gaze_deviation_threshold)
         {
/*          Restore any previous gazepoints that were temporarily left      */
/*          out of the fixation.                                            */
            RestoreOutPoints();

/*          Update the present fixation hypothesis, and check if there      */
/*          are enough samples to declare that the eye is fixating.         */
            UpdatePresFixation(x_gaze, y_gaze, minimum_fix_samples);
         }

/*   C2  Otherwise, if the point is outside the present fixation region,    */
         else   /* if (fPresDr > gaze_deviation_threshold) */
         {
/*          Increment the number of gazepoint samples outside the           */
/*          present fix.                                                    */
            iNPresOut++;

/*      D1  If the present fixation is finished, i.e. if there have         */
/*          been minimum_fix_samples since the gazepoint last matched       */
/*          the present fixation, and the present fixation is long          */
/*          enough to count as a real fixation,                             */
//would this be the place to cut the delay and call the stop?
            if (((short)(lCallCount - lPresFixEndCount) >= 2) &&
                 //minimum_fix_samples) &&
                (iNPresFixSamples >= minimum_fix_samples))
            {
/*             Declare the present fixation to be completed, move the       */
/*             present fixation to the prior, move the new fixation to      */
/*             the present, and check if the new (now present) fixation     */
/*             has enough points for the eye to be declared to be fixating. */
               DeclareCompletedFixation(minimum_fix_samples);

/*             Compute the deviation of the gazepoint from the now          */
/*             present fixation.                                            */
               CalcGazeDeviationFromPresFix(x_gaze, y_gaze);

/*         E1  If the gazepoint is within the now present fixation region,  */
               if (fPresDr <= gaze_deviation_threshold)
               {
/*                Update the present fixation data, and check if there      */
/*                are enough samples to declare that the eye is fixating.   */
                  UpdatePresFixation(x_gaze, y_gaze, minimum_fix_samples);
               }

/*         E2  Otherwise, if the gazepoint is outside the now present       */
/*             fixation,                                                    */
               else   /* if (fPresDr > gaze_deviation_threshold) */
               {
/*                Start a new fixation at the gazepoint.                    */
                  StartNewFixAtGazepoint(x_gaze, y_gaze);
               }
            }

/*      D2  Otherwise, if the present fixation is not finished,             */
            else
            {
/*         F1  If there is a new fixation hypothesis,                       */
               if (n_new_fix_samples > 0)
               {
/*                Compute the deviation of the gazepoint from the new       */
/*                fixation.                                                 */
                  CalcGazeDeviationFromNewFix(x_gaze, y_gaze);

/*            G1  If the new point falls within the new fix,                */
                  if (new_dr <= gaze_deviation_threshold)
                  {
/*                   Update the new fixation hypothesis.                    */
                     UpdateNewFixation(x_gaze, y_gaze);

/*               H.  If there are now enough points in the new fix          */
/*                   to declare it a real fix,                              */
                     if (n_new_fix_samples == minimum_fix_samples)
                     {
/*                      Drop the present fixation data, move the new        */
/*                      new fixation into the present fixation, and see     */
/*                      if the new (now present) fixation has enough        */
/*                      points to declare the eye to be fixating.           */
                        MoveNewFixToPresFix(minimum_fix_samples);
                     }
                  }

/*            G2  Otherwise, if the point is outside the new fix,           */
                  else   /* if (new_dr <= gaze_deviation_threshold) */
                  {
/*                   Start the new fixation at the new gazepoint.           */
                     StartNewFixAtGazepoint(x_gaze, y_gaze);
                  }
               }

/*         F2  Otherwise, If there is not a new fix,                        */
               else   /* if (n_new_fix_counts == 0) */
               {
/*                Start the new fixation at the gazepoint.                  */
                  StartNewFixAtGazepoint(x_gaze, y_gaze);
               }
            }
         }
      }

/* B2 Otherwise, if there is not a present fixation,                        */
      else  /* if (iNPresFixSamples = 0) */
      {
/*       Start the present fixation at the gazepoint and reset the          */
/*       new fixation.                                                      */
         StartPresFixAtGazepoint(x_gaze, y_gaze);
      }
   }

/*- - - - - - - - - - - - - Process Untracked Eye  - - - - - - - - - - - - -*/

/* A2 Otherwise, if the eye's gazepoint was not successfully measured       */
/* this sample,                                                             */
   else   /*  if (gazepoint_found == FALSE)  */
   {
/*    Increment the number of successive samples with no eye found.         */
      iNNoEyeFound++;

/*I   If it has been min-fix-samples since the last sample in the           */
/*    present fixation,                                                     */
      if ((short)(lCallCount - lPresFixEndCount) >= minimum_fix_samples)
      {
/*   J   If there had been a fixation prior to loosing track of the eye,    */
         if (iNPresFixSamples >= minimum_fix_samples)
         {
/*          Declare the present fixation to be completed, move the          */
/*          present fixation to the prior, move the new fixation to         */
/*          the present, and check if the new (now present) fixation        */
/*          has enough points for the eye to be declared to be fixating.    */
            DeclareCompletedFixation(minimum_fix_samples);
         }

/*       Reset the present fixation data.                                   */
         ResetPresFixation();
      }
   }

/*---------------------------- Pass Data Back ------------------------------*/

/* Pass the delayed gazepoint data, with the relevant saccade/fixation      */
/* data, back to the calling function.                                      */
   *ptr_x_gaze_delayed          = x_gaze_ring[iRingIndexDelay];
   *ptr_y_gaze_delayed          = y_gaze_ring[iRingIndexDelay];
   *ptr_gazepoint_found_delayed = gaze_found_ring[iRingIndexDelay];
   *ptr_x_fix_delayed           = x_fix_ring[iRingIndexDelay];
   *ptr_y_fix_delayed           = y_fix_ring[iRingIndexDelay];
   *ptr_gaze_deviation_delayed  = gaze_deviation_ring[iRingIndexDelay];
   *ptr_saccade_duration_delayed= sac_duration_ring[iRingIndexDelay];
   *ptr_fix_duration_delayed    = fix_duration_ring[iRingIndexDelay];

/* Return the eye motion/fixation state for the delayed point.              */

   return(eye_motion_state[iRingIndexDelay]);
}
/****************************************************************************/
void ResetPresFixation(void)

/* This function resets the present fixation, i.e. declares it nonexistent. */

{
   lPresFixStartCount = 0;
   lPresFixEndCount   = 0;
   iNPresFixSamples   = 0;
   fXPresFixSum       = 0.0F;
   fYPresFixSum       = 0.0F;
   fXPresFix          = 0.0F;
   fYPresFix          = 0.0F;
   iNPresOut          = 0;
}
/****************************************************************************/
void ResetNewFixation(void)

/* This function resets the new fixation, i.e. declares it nonexistent.     */

{
   new_fix_start_count = 0;
   new_fix_end_count   = 0;
   n_new_fix_samples   = 0;
   x_new_fix_sum       = 0.0F;
   y_new_fix_sum       = 0.0F;
   x_new_fix           = 0.0F;
   y_new_fix           = 0.0F;
}
/****************************************************************************/
void StartPresFixAtGazepoint(float x_gaze, float y_gaze)

/* This function starts the present fixation at the argument gazepoint      */
/* and makes sure there is no new fixation hypothesis.                      */


{

/* Start the present fixation at the argument gazepoint.                    */
   iNPresFixSamples   = 1;
   fXPresFixSum       = x_gaze;
   fYPresFixSum       = y_gaze;
   fXPresFix           = x_gaze;
   fYPresFix           = y_gaze;
   lPresFixStartCount = lCallCount;
   lPresFixEndCount   = lCallCount;
   iNPresOut           = 0;

/* Make sure there is no new fixation.                                      */
   ResetNewFixation();
}
/****************************************************************************/
void StartNewFixAtGazepoint(float x_gaze, float y_gaze)

/* This function starts the new fixation at the argument gazepoint.         */

{
   n_new_fix_samples   = 1;
   x_new_fix_sum       = x_gaze;
   y_new_fix_sum       = y_gaze;
   x_new_fix           = x_gaze;
   y_new_fix           = y_gaze;
   new_fix_start_count = lCallCount;
   new_fix_end_count   = lCallCount;
}
/****************************************************************************/
void UpdatePresFixation(float x_gaze, float y_gaze, short minimum_fix_samples)

/* This function updates the present fixation with the argument gazepoint,  */
/* checks if there are enough samples to declare that the eye is now        */
/* fixating, and makes sure there is no hypothesis for a new fixation.      */

{
/* Update the present fixation with the argument gazepoint.                 */
   iNPresFixSamples++;
   fXPresFixSum    += x_gaze;
   fYPresFixSum    += y_gaze;
   fXPresFix         = fXPresFixSum / iNPresFixSamples;
   fYPresFix         = fYPresFixSum / iNPresFixSamples;
   lPresFixEndCount = lCallCount;
   iNPresOut         = 0;

/* Check if there are enough samples in the present fixation hypothesis     */
/* to declare that the eye is fixating.                                     */
   CheckIfFixating(minimum_fix_samples);

/* There is no hypothesis for a new fixation.                               */
   ResetNewFixation();
}
/****************************************************************************/
void UpdateNewFixation(float x_gaze, float y_gaze)

/* This function updates the new fixation with the argument gazepoint.      */

{
/* Update the new fixation with the argument gazepoint.                     */
   n_new_fix_samples++;
   x_new_fix_sum    += x_gaze;
   y_new_fix_sum    += y_gaze;
   x_new_fix         = x_new_fix_sum / n_new_fix_samples;
   y_new_fix         = y_new_fix_sum / n_new_fix_samples;
   new_fix_end_count = lCallCount;
}
/****************************************************************************/
void CalcGazeDeviationFromPresFix(float x_gaze, float y_gaze)

/* This function calculates the deviation of the gazepoint from the         */
/* present fixation location.                                               */

{
   float dx, dy;                   /* horizontal and vertical deviations    */

   dx = x_gaze - fXPresFix;
   dy = y_gaze - fYPresFix;
   fPresDr = (float)sqrt(dx * dx + dy * dy);

/* Put the deviation in the ring buffer for future reference.               */
   gaze_deviation_ring[iRingIndex] = fPresDr;
}
/****************************************************************************/
void CalcGazeDeviationFromNewFix(float x_gaze, float y_gaze)

/* This function calculate the deviation of the gazepoint from the new      */
/* fixation location.                                                       */

{
   float dx, dy;                   /* horizontal and vertical deviations    */

   dx = x_gaze - x_new_fix;
   dy = y_gaze - y_new_fix;
   new_dr = (float)sqrt(dx * dx + dy * dy);
}
/****************************************************************************/
void CheckIfFixating(short minimum_fix_samples)

/* This function checks to see whether there are enough samples in the      */
/* presently hypothesized fixation to declare that the eye is fixating      */
/* yet, and if there is a true fixation going on, it updates the ring       */
/* buffers to reflect the fixation.                                         */

{
   short  i, ii;                  /* dummy ring indices                       */

/* If there are enough samples for a fixation,                              */
   if (iNPresFixSamples >= minimum_fix_samples)
   {
/*    Declare the eye to be fixating.  Go back through the last             */
/*    minimum_fix_samples entries of the ring buffer making sure that all   */
/*    samples from the present fixation are marked as fixating, and set     */
/*    the entries with the newest estimate of the fixation location.        */
      for (i = 0; i < minimum_fix_samples; i++)
      {
         ii = iRingIndex - i;
         if (ii < 0) ii += RING_SIZE;

         eye_motion_state[ii] = FIXATING;
         x_fix_ring[ii] = fXPresFix;
         y_fix_ring[ii] = fYPresFix;

         sac_duration_ring[ii] = (short)(lPresFixStartCount -
                                       lPrevFixEndCount - 1);
         fix_duration_ring[ii] = (short)(lPresFixEndCount -
                                       lPresFixStartCount + 1 - i);
      }
   }
}
/****************************************************************************/
void MoveNewFixToPresFix(short minimum_fix_samples)

/* This function copies the new fixation data into the present fixation,    */
/* and resets the new fixation.                                             */

{
/* Move the new fixation to the present fixation.                           */
   iNPresFixSamples   = n_new_fix_samples;
   fXPresFixSum       = x_new_fix_sum;
   fYPresFixSum       = y_new_fix_sum;
   fXPresFix          = x_new_fix;
   fYPresFix          = y_new_fix;
   lPresFixStartCount = new_fix_start_count;
   lPresFixEndCount   = new_fix_end_count;
   iNPresOut          = 0;

/* Reset the new fixation.                                                  */
   ResetNewFixation();

/* Check if there are enough samples in the new (now present) fixation to   */
/* declare that the eye is fixating.                                        */
   CheckIfFixating(minimum_fix_samples);
}
/****************************************************************************/
void DeclareCompletedFixation(short minimum_fix_samples)

/* This function:                                                           */
/* a) declares the present fixation to be completed,                        */
/* b) moves the present fixation to the prior fixation,                     */
/* c) moves the new fixation, if any, to the present fixation, and          */

{

/* Declare the present fixation to be completed.                            */
   eye_motion_state[iRingIndexDelay] = FIXATION_COMPLETED;

/* Move the present fixation to the previous fixation.  This saves the      */
/* end time of the present fixation for later computation of the saccade    */
/* period between this and the next fixation.                               */
   lPrevFixEndCount = lPresFixEndCount;

/* Move the new fixation data, if any, to the present fixation, reset       */
/* the new fixation, and check if there are enough samples in the new       */
/* (now present) fixation to declare that the eye is fixating.              */
   MoveNewFixToPresFix(minimum_fix_samples);
}
/****************************************************************************/
void RestoreOutPoints(void)

/* This function restores any previous gazepoints that were left out of     */
/* the fixation and are now known to be part of the present fixation.       */

{
   short  i, ii;                  /* dummy ring indices                       */

/* If there were some previous points that temporarily went out of the      */
/* fixation region,                                                         */
   if (iNPresOut > 0)
   {
/*    Undo the hypothesis that they were outside the fixation and declare   */
/*    them now to be part of the fix.                                       */
      for (i = 1; i <= iNPresOut; i++)
      {
         ii = iRingIndex - i;
         if (ii < 0) ii += RING_SIZE;

         if (gaze_found_ring[ii] == TRUE)
         {
            iNPresFixSamples++;
            fXPresFixSum    += x_gaze_ring[ii];
            fYPresFixSum    += y_gaze_ring[ii];
            eye_motion_state[ii] = FIXATING;
         }
      }

/*    Set the number of "out" points to be zero.                            */
      iNPresOut = 0;
   }
}
//this changes the state in the ringbuffer of the last minfixsamples. 
//Necessary in order to declare a fixation ended after only two samples.
void resetState(void){
short i, ii;
 for (i = 0; i < myminimum_fix_samples; i++)
      {
         ii = iRingIndex - i;
         if (ii < 0) ii += RING_SIZE;

         eye_motion_state[ii] = MOVING;
		}
}
/****************************************************************************/



	

	
